package org.openmetadata.service.resources.governance;

import io.swagger.v3.oas.annotations.ExternalDocumentation;
import io.swagger.v3.oas.annotations.Operation;
import io.swagger.v3.oas.annotations.Parameter;
import io.swagger.v3.oas.annotations.media.Content;
import io.swagger.v3.oas.annotations.media.ExampleObject;
import io.swagger.v3.oas.annotations.media.Schema;
import io.swagger.v3.oas.annotations.parameters.RequestBody;
import io.swagger.v3.oas.annotations.responses.ApiResponse;
import io.swagger.v3.oas.annotations.tags.Tag;
import java.io.IOException;
import java.util.UUID;
import javax.json.JsonPatch;
import javax.validation.Valid;
import javax.validation.constraints.Max;
import javax.validation.constraints.Min;
import javax.ws.rs.Consumes;
import javax.ws.rs.DELETE;
import javax.ws.rs.DefaultValue;
import javax.ws.rs.GET;
import javax.ws.rs.PATCH;
import javax.ws.rs.POST;
import javax.ws.rs.PUT;
import javax.ws.rs.Path;
import javax.ws.rs.PathParam;
import javax.ws.rs.Produces;
import javax.ws.rs.QueryParam;
import javax.ws.rs.core.Context;
import javax.ws.rs.core.MediaType;
import javax.ws.rs.core.Response;
import javax.ws.rs.core.SecurityContext;
import javax.ws.rs.core.UriInfo;
import org.openmetadata.schema.api.data.RestoreEntity;
import org.openmetadata.schema.api.governance.CreateWorkflowDefinition;
import org.openmetadata.schema.governance.workflows.WorkflowDefinition;
import org.openmetadata.schema.type.EntityHistory;
import org.openmetadata.schema.type.Include;
import org.openmetadata.service.Entity;
import org.openmetadata.service.OpenMetadataApplicationConfig;
import org.openmetadata.service.governance.workflows.Workflow;
import org.openmetadata.service.governance.workflows.WorkflowHandler;
import org.openmetadata.service.jdbi3.ListFilter;
import org.openmetadata.service.jdbi3.WorkflowDefinitionRepository;
import org.openmetadata.service.limits.Limits;
import org.openmetadata.service.resources.Collection;
import org.openmetadata.service.resources.EntityResource;
import org.openmetadata.service.security.Authorizer;
import org.openmetadata.service.util.EntityUtil;
import org.openmetadata.service.util.ResultList;

@Path("/v1/governance/workflowDefinitions")
@Tag(
    name = "Workflow Definitions",
    description =
        "A `Workflow Definition` is a configured workflow setup for a given governance task.")
@Produces(MediaType.APPLICATION_JSON)
@Consumes(MediaType.APPLICATION_JSON)
@Collection(name = "governanceWorkflows")
public class WorkflowDefinitionResource
    extends EntityResource<WorkflowDefinition, WorkflowDefinitionRepository> {
  public static final String COLLECTION_PATH = "v1/governance/workflowDefinitions/";
  static final String FIELDS = "owners";
  private final WorkflowDefinitionMapper mapper = new WorkflowDefinitionMapper();

  public WorkflowDefinitionResource(Authorizer authorizer, Limits limits) {
    super(Entity.WORKFLOW_DEFINITION, authorizer, limits);
  }

  public static class WorkflowDefinitionList extends ResultList<WorkflowDefinition> {
    /* Required for serde */
  }

  @Override
  public void initialize(OpenMetadataApplicationConfig config) throws IOException {
    repository.initSeedDataFromResources();
  }

  @GET
  @Operation(
      operationId = "listWorkflowDefinitions",
      summary = "List Workflow Definitions",
      description =
          "Get a list of Workflow Definitions. Use `fields` parameter to get only necessary fields."
              + " Use cursor-based pagination tom limit the number of entries in the list"
              + " using `limit` and `before` or `after` query params.",
      responses = {
        @ApiResponse(
            responseCode = "200",
            description = "List of Workflow Definitions",
            content =
                @Content(
                    mediaType = "application/json",
                    schema =
                        @Schema(
                            implementation =
                                WorkflowDefinitionResource.WorkflowDefinitionList.class)))
      })
  public ResultList<WorkflowDefinition> list(
      @Context UriInfo uriInfo,
      @Context SecurityContext securityContext,
      @Parameter(
              description = "Fields requested in the returned resource",
              schema = @Schema(type = "string", example = FIELDS))
          @QueryParam("fields")
          String fieldsParam,
      @Parameter(description = "Limit the number of Workflow Definitions returned. Default = 10)")
          @DefaultValue("10")
          @QueryParam("limit")
          @Min(0)
          @Max(1000000)
          int limitParam,
      @Parameter(description = "Returns the list of Workflow Definitions before this cursor")
          @QueryParam("before")
          String before,
      @Parameter(description = "Returns the list of Workflow Definitions after this cursor")
          @QueryParam("after")
          String after,
      @Parameter(description = "Include all, deleted or non-deleted entities.")
          @QueryParam("include")
          @DefaultValue("non-deleted")
          Include include) {
    ListFilter filter = new ListFilter(include);
    return super.listInternal(
        uriInfo, securityContext, fieldsParam, filter, limitParam, before, after);
  }

  @GET
  @Path("/{id}/versions")
  @Operation(
      operationId = "listAllWorkflowDefinitionVersion",
      summary = "List Workflow Definition versions",
      description = "Get a list of all the versions of a Workflow Definition identified by `Id`",
      responses = {
        @ApiResponse(
            responseCode = "200",
            description = "List of Workflow Definition versions",
            content =
                @Content(
                    mediaType = "application/json",
                    schema = @Schema(implementation = EntityHistory.class)))
      })
  public EntityHistory listVersions(
      @Context UriInfo uriInfo,
      @Context SecurityContext securityContext,
      @Parameter(description = "Id of the Workflow Definition", schema = @Schema(type = "UUID"))
          @PathParam("id")
          UUID id) {
    return super.listVersionsInternal(securityContext, id);
  }

  @GET
  @Path("/{id}")
  @Operation(
      operationId = "getWorkflowDefinitionByID",
      summary = "Get a Workflow Definition by Id",
      description = "Get a Workflow Definition by `Id`.",
      responses = {
        @ApiResponse(
            responseCode = "200",
            description = "The Workflow Definition",
            content =
                @Content(
                    mediaType = "application/json",
                    schema = @Schema(implementation = WorkflowDefinition.class))),
        @ApiResponse(
            responseCode = "404",
            description = "Workflow Definition for instance {id} is not found")
      })
  public WorkflowDefinition get(
      @Context UriInfo uriInfo,
      @Parameter(description = "Id of the Workflow Definition", schema = @Schema(type = "UUID"))
          @PathParam("id")
          UUID id,
      @Context SecurityContext securityContext,
      @Parameter(
              description = "Fields requested in the returned resource",
              schema = @Schema(type = "string", example = FIELDS))
          @QueryParam("fields")
          String fieldsParam,
      @Parameter(
              description = "Include all, deleted, or non-deleted entities.",
              schema = @Schema(implementation = Include.class))
          @QueryParam("include")
          @DefaultValue("non-deleted")
          Include include) {
    return getInternal(uriInfo, securityContext, id, fieldsParam, include);
  }

  @POST
  @Path("/{id}/redeploy")
  @Operation(
      operationId = "getWorkflowDefinitionByID",
      summary = "Get a Workflow Definition by Id",
      description = "Get a Workflow Definition by `Id`.",
      responses = {
        @ApiResponse(
            responseCode = "200",
            description = "The Workflow Definition",
            content =
                @Content(
                    mediaType = "application/json",
                    schema = @Schema(implementation = WorkflowDefinition.class)))
      })
  public Response redeploy(
      @Context UriInfo uriInfo,
      @Parameter(description = "Id of the Workflow Definition", schema = @Schema(type = "UUID"))
          @PathParam("id")
          UUID id,
      @Context SecurityContext securityContext) {
    WorkflowDefinition wd =
        repository.get(
            uriInfo,
            id,
            new EntityUtil.Fields(repository.getAllowedFields()),
            Include.NON_DELETED,
            false);
    WorkflowHandler.getInstance().deleteWorkflowDefinition(wd);
    WorkflowHandler.getInstance().deploy(new Workflow(wd));
    return Response.status(Response.Status.OK).entity("Workflow Redeployed").build();
  }

  @GET
  @Path("/name/{fqn}")
  @Operation(
      operationId = "getWorkflowDefinitionByFQN",
      summary = "Get a Workflow Definition by fully qualified name",
      description = "Get a Workflow Definition by `fullyQualifiedName`.",
      responses = {
        @ApiResponse(
            responseCode = "200",
            description = "The Workflow Definition",
            content =
                @Content(
                    mediaType = "application/json",
                    schema = @Schema(implementation = WorkflowDefinition.class))),
        @ApiResponse(
            responseCode = "404",
            description = "Workflow Definition for instance {fqn} is not found")
      })
  public WorkflowDefinition getByName(
      @Context UriInfo uriInfo,
      @Parameter(
              description = "Fully qualified name of the Workflow Definition",
              schema = @Schema(type = "string"))
          @PathParam("fqn")
          String fqn,
      @Context SecurityContext securityContext,
      @Parameter(
              description = "Fields requested in the returned resource",
              schema = @Schema(type = "string", example = FIELDS))
          @QueryParam("fields")
          String fieldsParam,
      @Parameter(
              description = "Include all, deleted, or non-deleted entities.",
              schema = @Schema(implementation = Include.class))
          @QueryParam("include")
          @DefaultValue("non-deleted")
          Include include) {
    return getByNameInternal(uriInfo, securityContext, fqn, fieldsParam, include);
  }

  @GET
  @Path("/{id}/versions/{version}")
  @Operation(
      operationId = "getSpecificWorkflowDefinitionVersion",
      summary = "Get a version of the Workflow Definition",
      description = "Get a version of the Workflow Definition by given `Id`",
      responses = {
        @ApiResponse(
            responseCode = "200",
            description = "database",
            content =
                @Content(
                    mediaType = "application/json",
                    schema = @Schema(implementation = WorkflowDefinition.class))),
        @ApiResponse(
            responseCode = "404",
            description =
                "Workflow Definition for instance {id} and version {version} is not found")
      })
  public WorkflowDefinition getVersion(
      @Context UriInfo uriInfo,
      @Context SecurityContext securityContext,
      @Parameter(description = "Id of the Workflow Definition", schema = @Schema(type = "UUID"))
          @PathParam("id")
          UUID id,
      @Parameter(
              description = "WorkflowDefinition version number in the form `major`.`minor`",
              schema = @Schema(type = "string", example = "0.1 or 1.1"))
          @PathParam("version")
          String version) {
    return super.getVersionInternal(securityContext, id, version);
  }

  @POST
  @Operation(
      operationId = "createWorkflowDefinition",
      summary = "Create a Workflow Definition",
      description = "Create a Workflow Definition.",
      responses = {
        @ApiResponse(
            responseCode = "200",
            description = "The Workflow Definition",
            content =
                @Content(
                    mediaType = "application/json",
                    schema = @Schema(implementation = WorkflowDefinition.class))),
        @ApiResponse(responseCode = "400", description = "Bad request")
      })
  public Response create(
      @Context UriInfo uriInfo,
      @Context SecurityContext securityContext,
      @Valid CreateWorkflowDefinition create) {
    WorkflowDefinition workflowDefinition =
        mapper.createToEntity(create, securityContext.getUserPrincipal().getName());
    return create(uriInfo, securityContext, workflowDefinition);
  }

  @PATCH
  @Path("/{id}")
  @Operation(
      operationId = "patchWorkflowDefinition",
      summary = "Update a Workflow Definition by Id",
      description = "Update an existing Workflow Definition using JsonPatch.",
      externalDocs =
          @ExternalDocumentation(
              description = "JsonPatch RFC",
              url = "https://tools.ietf.org/html/rfc6902"))
  @Consumes(MediaType.APPLICATION_JSON_PATCH_JSON)
  public Response patch(
      @Context UriInfo uriInfo,
      @Context SecurityContext securityContext,
      @Parameter(description = "Id of the Workflow Definition", schema = @Schema(type = "UUID"))
          @PathParam("id")
          UUID id,
      @RequestBody(
              description = "JsonPatch with array of operations",
              content =
                  @Content(
                      mediaType = MediaType.APPLICATION_JSON_PATCH_JSON,
                      examples = {
                        @ExampleObject("[{op:remove, path:/a},{op:add, path: /b, value: val}]")
                      }))
          JsonPatch patch) {
    return patchInternal(uriInfo, securityContext, id, patch);
  }

  @PATCH
  @Path("/name/{fqn}")
  @Operation(
      operationId = "patchWorkflowDefinition",
      summary = "Update a Workflow Definition by name.",
      description = "Update an existing Workflow Definition using JsonPatch.",
      externalDocs =
          @ExternalDocumentation(
              description = "JsonPatch RFC",
              url = "https://tools.ietf.org/html/rfc6902"))
  @Consumes(MediaType.APPLICATION_JSON_PATCH_JSON)
  public Response patch(
      @Context UriInfo uriInfo,
      @Context SecurityContext securityContext,
      @Parameter(description = "Name of the Workflow Definition", schema = @Schema(type = "string"))
          @PathParam("fqn")
          String fqn,
      @RequestBody(
              description = "JsonPatch with array of operations",
              content =
                  @Content(
                      mediaType = MediaType.APPLICATION_JSON_PATCH_JSON,
                      examples = {
                        @ExampleObject("[{op:remove, path:/a},{op:add, path: /b, value: val}]")
                      }))
          JsonPatch patch) {
    return patchInternal(uriInfo, securityContext, fqn, patch);
  }

  @PUT
  @Operation(
      operationId = "createOrUpdateWorkflowDefinition",
      summary = "Create or update Workflow Definition",
      description =
          "Create a Workflow Definition, if it does not exist or update an existing Workflow Definition.",
      responses = {
        @ApiResponse(
            responseCode = "200",
            description = "The updated Workflow Definition",
            content =
                @Content(
                    mediaType = "application/json",
                    schema = @Schema(implementation = WorkflowDefinition.class)))
      })
  public Response createOrUpdate(
      @Context UriInfo uriInfo,
      @Context SecurityContext securityContext,
      @Valid CreateWorkflowDefinition create) {
    WorkflowDefinition workflowDefinition =
        mapper.createToEntity(create, securityContext.getUserPrincipal().getName());
    return createOrUpdate(uriInfo, securityContext, workflowDefinition);
  }

  @DELETE
  @Path("/{id}")
  @Operation(
      operationId = "deleteWorkflowDefinition",
      summary = "Delete a Workflow Definition by Id",
      description = "Delete a Workflow Definition by `Id`.",
      responses = {
        @ApiResponse(responseCode = "200", description = "OK"),
        @ApiResponse(
            responseCode = "404",
            description = "Workflow Definition for instance {id} is not found")
      })
  public Response delete(
      @Context UriInfo uriInfo,
      @Context SecurityContext securityContext,
      @Parameter(
              description = "Recursively delete this entity and it's children. (default `false`)")
          @DefaultValue("false")
          @QueryParam("recursive")
          boolean recursive,
      @Parameter(description = "Hard delete the entity. (default = `false`)")
          @QueryParam("hardDelete")
          @DefaultValue("false")
          boolean hardDelete,
      @Parameter(description = "Id of the Workflow Definition", schema = @Schema(type = "UUID"))
          @PathParam("id")
          UUID id) {
    return delete(uriInfo, securityContext, id, recursive, hardDelete);
  }

  @DELETE
  @Path("/name/{fqn}")
  @Operation(
      operationId = "deleteWorkflowDefinitionByFQN",
      summary = "Delete a Workflow Definition by fully qualified name",
      description = "Delete a Workflow Definition by `fullyQualifiedName`.",
      responses = {
        @ApiResponse(responseCode = "200", description = "OK"),
        @ApiResponse(
            responseCode = "404",
            description = "Workflow Definition for instance {fqn} is not found")
      })
  public Response delete(
      @Context UriInfo uriInfo,
      @Context SecurityContext securityContext,
      @Parameter(description = "Hard delete the entity. (default = `false`)")
          @QueryParam("hardDelete")
          @DefaultValue("false")
          boolean hardDelete,
      @Parameter(
              description = "Recursively delete this entity and it's children. (default `false`)")
          @QueryParam("recursive")
          @DefaultValue("false")
          boolean recursive,
      @Parameter(
              description = "Fully qualified name of the Workflow Definition",
              schema = @Schema(type = "string"))
          @PathParam("fqn")
          String fqn) {
    return deleteByName(uriInfo, securityContext, fqn, recursive, hardDelete);
  }

  @PUT
  @Path("/restore")
  @Operation(
      operationId = "restore",
      summary = "Restore a soft deleted Workflow Definition.",
      description = "Restore a soft deleted Workflow Definition.",
      responses = {
        @ApiResponse(
            responseCode = "200",
            description = "Successfully restored the Workflow Definition. ",
            content =
                @Content(
                    mediaType = "application/json",
                    schema = @Schema(implementation = WorkflowDefinition.class)))
      })
  public Response restore(
      @Context UriInfo uriInfo,
      @Context SecurityContext securityContext,
      @Valid RestoreEntity restore) {
    return restoreEntity(uriInfo, securityContext, restore.getId());
  }

  @POST
  @Path("/name/{fqn}/trigger")
  @Operation(
      operationId = "triggerWorkflow",
      summary = "Start a new instance of a Workflow Definition",
      description = "Start a new instance of a Workflow Definition.",
      responses = {
        @ApiResponse(
            responseCode = "200",
            description = "Workflow trigger status code",
            content = @Content(mediaType = "application/json")),
        @ApiResponse(
            responseCode = "404",
            description = "Workflow Definition named '{fqn}' is not found")
      })
  public Response trigger(
      @Context UriInfo uriInfo,
      @Context SecurityContext securityContext,
      @Parameter(description = "Name of the Workflow Definition", schema = @Schema(type = "string"))
          @PathParam("fqn")
          String fqn) {
    boolean triggerResponse = WorkflowHandler.getInstance().triggerWorkflow(fqn);
    if (triggerResponse) {
      return Response.status(Response.Status.OK).entity("Workflow Triggered").build();
    } else {
      return Response.status(Response.Status.NOT_FOUND).entity(fqn).build();
    }
  }
}
